︠c0247583-3f22-42d1-b88e-c3206fbe5f0es︠
import numpy
import itertools

############################################################################################################################
# For all these procedures we are working with elliptic curve (EC) groups over Z_p where p is a prime number larger than 3.#
# An Elliptic Curve group will be represented by a triple [A,B,p] where                                                    #
#      (1) p is a prime number larger than 3                                                                               #
#      (2) A is the coefficient of x and                                                                                   #
#      (3) B is the constant term in  y^2 = x^3 + A*x + B mod p.                                                           #
#                                                                                                                          #
# A point in the group will be represented as a pair [x,y] where                                                           #
#      (1)  x is the x-coordinate of a solution                                                                            #
#      (2)  y is the y-coordinate of a solution                                                                            #
#      (3)  The identity will be represented by []                                                                         #
############################################################################################################################


def TonSh (a, Prime):
    if kronecker(a, Prime) == -1:
        pass
        print "{0} does not have a square root mod {1}".format(a, Prime)
        return None
    elif Prime % 4 == 3:
        x = power_mod(ZZ(a), ZZ((Prime+1)/4),ZZ(Prime))
        return(x)
    else:
        #################################################################################
        # Determine e so that Prime-1 = (2^e)*q for some odd number q                   # 
        #################################################################################

        e = ZZ(0)
        q = ZZ(Prime - 1)
#        for j in range(1, Prime):
        for j in itertools.count(1):
            if q % 2 == 0:
                e = ZZ(e + 1)
                q = ZZ(q / 2)
            else:
                break
        for i in range(1, 101):
            n = i
            if kronecker(n, Prime) == -1:
                break
        z = power_mod(ZZ(n),ZZ(q),ZZ(Prime))
        y = ZZ(z)
        r = ZZ(e)
        x = power_mod(ZZ(a),ZZ( (q-1)/2),ZZ( Prime))
        b = (a * x ** 2) % Prime
        x = (a * x) % Prime
        #for i in range(1, e + 1):
        for i in itertools.count(1):
            if b == 1:
                break
            else:
                for m in itertools.count(1):
                    t = power_mod(ZZ(b), ZZ(2^m) ,  ZZ(Prime))
                    if t == 1:
                        mm = m
                        break
                t = power_mod(ZZ(y), ZZ(2^(r - mm - 1)),ZZ(Prime))
                y = power_mod(ZZ(t), ZZ(2),ZZ(Prime))
                r = mm
                x = (x * t) % Prime
                b = (b * y) % Prime
        return(x)


︡c87b844f-8c81-4541-a459-0d73197446d9︡{"stderr":"Error in lines 3-44\nTraceback (most recent call last):\n  File \"/cocalc/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 1231, in execute\n    compile(block + '\\n',\n  File \"<string>\", line 4\n    print \"{0} does not have a square root mod {1}\".format(a, Prime)\n          ^\nSyntaxError: invalid syntax\n"}︡{"done":true}
︠236c4656-e579-433e-8e97-d3abdb067738i︠
%html
<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between "lowerbound" and "upperbound". The lowerbound and the upperbound are positive random numbers. 
︡eefda683-06b5-4a0c-89e8-4f730397080d︡︡{"done":true,"html":"<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between \"lowerbound\" and \"upperbound\". The lowerbound and the upperbound are positive random numbers."}
︠e6c93c60-6ba8-4bf5-b5db-7557b6152d72︠
import itertools
def ECSearch (lowerbound, upperbound, Group):
    p = Group[2]
    a = Group[0]
    b = Group[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print "This is not an elliptic curve. "
    else:
        for j in itertools.count(lowerbound):
            if j==lowerbound-1 or j>upperbound or j>upperbound:
                return "not found"
            j=j%p
            #print "test "+str(j)
            #print (kronecker(j ** 3 + a * j + b, p) )
            if kronecker(j ** 3 + a * j + b, p) == 1:
                x = (j ** 3 + a * j + b) % p
                var('z')
                #print("stuck here")
                #print ("Modsqrt({0}, {1})".format(x,p))
                #L = solve_mod(z ** 2 == x, p)
                L=TonSh(x,p)
                y = L
                #if y>p/2:
                #    y=p/2-(y-p/2)
                #y = L[0][0]
                return([j,y])
︡782832e4-25ac-4c89-b4b4-62b28fb09932︡︡{"done":true}
︠8453f9ac-1f94-49c6-a3e1-ae68a4533b47︠

︡45b794d2-5174-4c46-8083-4ebe4db97629︡︡{"done":true}
︠de2a39a2-cab7-4b4f-ac69-d3982d5e189f︠

import sympy.ntheory
def FermatGreat (N):
    if sympy.ntheory.isprime(N):
        if N % 4 == 1:
            Z = two_squares(N)
            #print(Z)
            return(Z)
        else:
            print "{0} is prime, but {1} mod 4 = 3".format(N, N)
    else:
        print "{0} is not prime\n".format(N)

︡2471b9bd-9eea-41e5-9f2f-6b4563c3bd9b︡︡{"done":true}
︠79629385-1c87-4dad-a38f-2d74d025669ei︠

%html
<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation. 
︡2358f900-5d21-4047-bac4-3c125f016827︡︡{"done":true,"html":"<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation."}
︠f3b0703d-ea25-4929-a253-57c752ece319︠
︡cf05e58a-e130-4702-a96d-7c67c4b98f29︡︡{"done":true}
︠d6ba2d47-db9c-4531-b62c-32bdefe2aa8d︠
def ECAdd (Point1, Point2, Group):
    
    a = Group[0]
    b = Group[1]
    p = Group[2]
    '''
    #This could be an optimization for code readability
    E = EllipticCurve(GF(p),[a,b])
    p1 = E(Point1[0],Point1[1])
    p2 = E(Point2[0],Point2[1])
    toReturn = p1+p2
    return [toReturn[0],toReturn[1]]
    '''
    if(Point1!=[]):
        x1 = Point1[0]
        y1 = Point1[1]
    if(Point2!=[]):
        x2 = Point2[0]
        y2 = Point2[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print "This is not an elliptic curve. "
    elif Point1 != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
        print "Point 1 is not on the elliptic curve. \n"
    elif Point2 != [] and y2 ^ 2 % p != (x2 ^ 3 + a * x2 + b) % p:
        print "Point 2 is not on the elliptic curve. \n"
    else:
        if Point1 == []:
            R = Point2
        elif Point2 == []:
            R = Point1
        else:
            if x1 == x2 and 0 == (y1 + y2) % p:
                R = []
            elif x1 == x2 and y1 == y2:
                R = ECDouble(Point1, Group)
            else:
                s = ((y1 - y2) / (x1 - x2)) % p
                x = (s ** 2 - x1 - x2) % p
                y = (s * (x1 - x) - y1) % p
                R = [x,y]
    return(R)
︡84a194e5-ce31-48bb-992e-c0dd62d2fec2︡︡{"done":true}
︠795080ab-d669-44e3-bd82-7d5e0cdd17ae︠
︡9908ea8f-1ef8-4292-b6fc-b81886b1895c︡︡{"done":true}
︠2c75fed5-4477-4f1b-85e7-a17925d96a6c︠
def ECDouble (Point, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if Point != []:
        x1 = Point[0]
        y1 = Point[1]
        if (4 * a ** 3 + 27 * b ** 2) % p == 0:
            print "This is not an elliptic curve. "
        elif Point != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
            print "The point to double is not on the elliptic curve. "
        elif y1 == 0:
            R = []
        else:
            s = mod((3 * x1 ** 2 + a) / y1 / 2, p)
            x = (s ** 2 - 2 * x1) % p
            y = (s * (x1 - x) - y1) % p
            R = [x,y]
    else:
        R = []
    return(R)
︡7a937665-f819-4a68-bd6e-218219e797e3︡︡{"done":true}
︠1e3370a5-ca4b-4c48-9b2f-d5f541a9c1f2︠
︡2941e271-89b6-4545-8c3b-1ebafb567cd9︡︡{"done":true}
︠36c514f0-6055-4771-9023-dd6ec4723281︠

def ECInverse (Point, Group):
    if Point == []:
        return(Point)
    else:
        p = Group[2]
        x = Point[0]
        y = Point[1]
        return([x,(p - y) % p])
︡b4edc0a2-7507-4bda-af05-be8397e2e90b︡︡{"done":true}
︠248b5919-f1b5-4029-ace3-e1a2a98d2caai︠
%html
<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation.
︡18b4f37e-91cf-4637-ba84-cac38e1450c1︡︡{"done":true,"html":"<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation."}
︠db90e0c4-4f08-428b-933d-8daa383a99d9︠
def ECTimes (Point, scalar, Group):
    ECIdentity=[]
    if Point == ECIdentity or scalar == 0:
        return(ECIdentity)
    else:
        E=EllipticCurve(GF(Group[2]),[Group[0],Group[1]])
        Eid = E(0,1,0)
        EPoint = E(Point[0],Point[1])
        #print type(EPoint)
        cgret = scalar*EPoint
        if cgret == Eid:
            print "ectimes id"
            return []
        #print cgret[0]
        return([ZZ(cgret[0]),ZZ(cgret[1])])

︡688bd039-a706-46c3-a886-7da64bd1a334︡︡{"done":true}
︠e60120fa-19d1-4724-8b25-5767f75ace77︠

︡2431db42-f133-4499-bb23-b882f9bd0b45︡
︠e7a25eb8-4d90-42cd-880c-cf066e34bc2c︠

︡1116d32c-c84d-45a4-8628-185b57412fe8︡
︠1afc24b6-680c-47d8-a6bc-c0cbf23a0a22︠
def findptOrder(point,group):
    E = EllipticCurve(GF(group[2]),[group[0],group[1]])
    Ep = E(point[0],point[1])
    return Ep.additive_order()

def Dirichlet (q):
    j=0

    #for j in numpy.arange(0.10e1, infinity + 0.10e1, 0.10e1):
    while True:
        j+=1
        p = 2 * j * q + 1
        if is_prime(p) == True:
            return(p)


︡ba913e00-67b8-416a-afd4-2e68cc4614a6︡︡{"done":true}︡{"done":true}
︠272fa4e5-98c4-44f3-b0e2-e12047b23bbbs︠
#Constructing El Gamal key resistant against Hellman Pohlig Silver attack
#Choose a large prime number q
q=next_prime(283743928792183019830289149837287463847938240218302198308921094832985743875834);
#Use the Dirichlet procedure to find a prime number p such that (p-1) has q as one of its prime factors
p=Dirichlet(q);
p
A=3476283764
B=0
G=[A,B,p]
k=factor(p+1);
k

︡ca4c99e5-9a2f-471f-b57b-d328e18ff02f︡{"stderr":"Error in lines 2-2\nTraceback (most recent call last):\n  File \"/cocalc/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 1230, in execute\n    exec(\n  File \"\", line 1, in <module>\nNameError: name 'Dirichlet' is not defined\n"}︡{"done":true}
︠0f699180-a79b-4dfa-b9ce-dd9c6449924f︠


︡37c1fef2-9481-403a-86e9-0bac0b6d72a2︡
︠258609d8-674e-4862-aab5-c3d5dd5e0b63︠

#Constructing a signature key that is resistant to the HPSonP attacks

q1 = next_prime(37493278489327498732487398498393879480898493287489732974921094832985743875834)
p1 = Dirichlet(q1)
fL1=list(factor(p1-1))
for test in fL1:
    print test
c11 = 1
c21 = 100
c31 = 374932784893274987324873984983938794808984932874897329749210948329857438758
x1 = crt([c11,c21,c31],[fL1[0][0],fL1[1][0],fL1[2][0]])
x1
g1=findLargePrimitiveRoot(10038748923784,p1)
g1
b1 = power_mod(g1,x1,p1)
b1
︡dbdf9a8d-ddfd-42b3-95fc-e3c91a6a4ccb︡︡{"stdout":"(2, 1)\n(113, 1)\n(37493278489327498732487398498393879480898493287489732974921094832985743875881, 1)\n","done":false}︡{"stdout":"8061429807990305502472115551139668027187985041743167486937784600040264790753173\n","done":false}︡{"stdout":"10038748923786\n","done":false}︡{"stdout":"6156089434541705029559508896034912822846473392122022710870417979600647684748946\n","done":false}︡{"done":true}
︠74ba88f9-e862-4388-be35-9f425a42adef︠
q2 = next_prime(89324793827492374982748732402938332894798230912830912921094832985743875834)
p2 = Dirichlet(q2)
fL2=list(factor(p2-1))
c12 = 1
c22 = 30
c32 = 8932479382749237498274873240293833289479823091283091292109483298574387616
x2 = crt([c12,c22,c32],[fL2[0][0],fL2[1][0],fL2[2][0]])
x2
g2=findLargePrimitiveRoot(10038748923784,p2)
g2
b2 = power_mod(g2,x2,p2)
b2
︡bcf9d1f8-76fa-44fd-9c19-999484c9867f︡︡{"stdout":"5100445727549814611514952620207778808292978985122645127794514963485975329249\n","done":false}︡{"stdout":"10038748923785\n","done":false}︡{"stdout":"946731983399710512555195230311934479072009381576909000168330179607867912495\n","done":false}︡{"done":true}
︠33450bee-43e0-40f9-a08a-0142d748d897︠

q3 = next_prime(898328932749832794872478923748923749873289473892743829749238742189723879294832985743875834)
p3 = Dirichlet(q3)
fL3=list(factor(p3-1))
c13 = 1
c23 = 50
c33 = 898328932749832794872478923748923749873289473892743829749238742189723879294832985743875859
x3 = crt([c13,c23,c33],[fL3[0][0],fL3[1][0],fL3[2][0]])
x3
︡b9471446-21bf-45b3-ae4f-8d92ca848951︡︡{"stdout":"67374669956237459615435919281169281240496710541955787231192905664229290947112473930790689425\n","done":false}︡{"done":true}










